

from collections import namedtuple

Signal = namedtuple('Signal', 'id content')
Signal.__new__.__defaults__ = (None,)


def make_signal(id):
    func = lambda content: Signal(id=id, content=content)
    func.__defaults__ = (None,)
    return func


ID_TASK_STOP = -1
ID_TASK_BUFF = 0
ID_BLOCK_FINISH = 1
ID_TASK_SLICE = 2
ID_TASK_FINISH = 3
ID_THREAD_END = 4
ID_TASK_PAUSE = 5
ID_TASK_EXCEPTION = 6
ID_TASK_OPEN = 7
ID_TASK_START = 8
ID_TASK_FAIL = 9
ID_EMPTY_RECV = 15

ID_CALLBACK_END = 10

ID_STOP_ALL = 16
ID_IDLE = 17

ID_URL_STATUS = 100

ID_NORMAL = 0
ID_CRASH = 1
ID_TIMEOUT = 2
ID_GAIERROR = 3
ID_UNKNOWN = 4
ID_SWITCH = 5
ID_WAIT = 6


SIGNAL_TASK_STOP = make_signal(ID_TASK_STOP)

SIGNAL_STOP_ALL = make_signal(ID_STOP_ALL)

SIGNAL_TASK_FINISH = make_signal(ID_TASK_FINISH)

SIGNAL_TASK_PAUSE = make_signal(ID_TASK_PAUSE)

SIGNAL_TASK_OPEN = make_signal(ID_TASK_OPEN)
SIGNAL_TASK_START = make_signal(ID_TASK_START)
SIGNAL_TASK_BUFF = make_signal(ID_TASK_BUFF)
SIGNAL_TASK_SLICE = make_signal(ID_TASK_SLICE)
SIGNAL_THREAD_END = make_signal(ID_THREAD_END)

SIGNAL_EMPTY_RECV = make_signal(ID_EMPTY_RECV)

SIGNAL_URL_STATUS = make_signal(ID_URL_STATUS)

SIGNAL_NORMAL = make_signal(ID_NORMAL)
SIGNAL_CRASH = make_signal(ID_CRASH)
SIGNAL_TIMEOUT = make_signal(ID_TIMEOUT)
SIGNAL_GAIERROR = make_signal(ID_GAIERROR)
SIGNAL_UNKNOWN = make_signal(ID_UNKNOWN)

SIGNAL_EXCEPTION = make_signal(ID_TASK_EXCEPTION)

SIGNAL_SWITCH = make_signal(ID_SWITCH)

SIGNAL_WAIT = make_signal(ID_WAIT)

SIGNAL_IDLE = make_signal(ID_IDLE)

SIGNAL_TASK_FAIL = make_signal(ID_TASK_FAIL)

SIGNAL_CALLBACK_END = make_signal(ID_CALLBACK_END)





from six.moves.http_client import HTTPSConnection as HTTPSConnection_origin, HTTPConnection

from nbdler.client.abstract_client import AbstractClient
from six.moves.urllib.error import URLError, HTTPError
from nbdler.exception import URLUnknownError, HTTP4XXError, URLConnectFailed
from nbdler.misc.signal import SIGNAL_TASK_BUFF, SIGNAL_TASK_SLICE, ID_TASK_SLICE, SIGNAL_THREAD_END, ID_TASK_PAUSE, \
    SIGNAL_GAIERROR, SIGNAL_CRASH, SIGNAL_TIMEOUT, SIGNAL_UNKNOWN, \
    SIGNAL_EMPTY_RECV, SIGNAL_URL_STATUS, SIGNAL_NORMAL, ID_SWITCH, ID_WAIT

from six.moves.queue import Queue
from nbdler.struct.misc import Slice, InitialResult, ClientException
import re
import ssl
import socket
from time import sleep

https_context = ssl._create_default_https_context()
HTTPSConnection = lambda **kwargs: HTTPSConnection_origin(context=https_context, **kwargs)


class HTTPClient(AbstractClient):
    __connected = False

    def __init__(self, source, progress):
        self.source = source
        self.progress = progress

        self._response = None
        self._connection = None

        self._callback = None
        self.__signal = Queue()

    def geturl(self):
        return self.source.url

    def _buffer(self, buff):
        if buff:
            self.progress.buffer(buff)
            self._callback.put_nowait(SIGNAL_TASK_BUFF(len(buff)))

    def _build_connection(self):
        if self.source.scheme == 'http':
            client_handler = HTTPConnection
        elif self.source.scheme == 'https':
            client_handler = HTTPSConnection
        else:
            raise ValueError('http client got an unexpected url protocol (%s).' %
                             self.source.scheme)

        conn = client_handler(host=self.source.hostname, port=self.source.port, timeout=5)

        req_range = (self.progress.begin + self.progress.go_inc, self.progress.end)
        path, headers = self.source.http_request_header(req_range)
        conn.request('GET', path, '', dict(headers))

        return conn

    def _conn_response(self, conn):
        res = conn.getresponse()
        sleep(0.01)
        if res.code in (301, 302, 303, 307):
            redurl = res.getheader('location', None)
            self.source.http_redirect(redurl)
            res.close()
            conn.sock.shutdown(socket.SHUT_RDWR)
            conn.close()
            conn = self._build_connection()
            return self._conn_response(conn)
        elif 400 <= res.code < 500:
            raise HTTP4XXError("(%d)URL: %s " % (res.code, self.source.url))
        elif res.code not in (200, 206):
            self._callback.put_nowait(
                SIGNAL_URL_STATUS(SIGNAL_UNKNOWN(self))
            )
            raise URLUnknownError()
        return conn, res

    def connect(self):
        conn = self._build_connection()
        conn, res = self._conn_response(conn)
        self._connection = conn
        self._response = res
        self.source.response(self.source.url, res.getheaders(), res.getcode(), res.length)
        self.__connected = True
        return res

    def retrieve(self):

        self._callback.put_nowait(
            SIGNAL_URL_STATUS(SIGNAL_NORMAL(ClientException(client=self, exception=None)))
        )
        self.progress.start()
        buff = b''
        while True:
            if not self.__signal.empty():
                signal = self.__signal.get()
                if signal.id == ID_TASK_SLICE:
                    putrange = self.progress.slice_check(signal.content)
                    self._callback.put_nowait(
                        SIGNAL_TASK_SLICE(Slice(client=HTTPClient, range=putrange))
                    )
                elif signal.id == ID_TASK_PAUSE:
                    self._buffer(buff)
                    self.__signal.task_done()
                    break
                elif signal.id == ID_WAIT:
                    sleep(signal.content)

                self.__signal.task_done()

            prv_len = len(buff)
            remain = self.progress.length - self.progress.go_inc
            try:
                if remain >= 8192:
                    buff += self._response.read(8192)
                elif remain > 0:
                    buff += self._response.read(remain)
                else:
                    self._buffer(buff)
                    break
            except (socket.gaierror, URLError, HTTPError, socket.timeout, Exception) as e:
                self._buffer(buff)
                self._exception_handler(e)
                break

            if len(buff) - prv_len == 0:
                self._buffer(buff)
                if self._response.chunked and not self._response.fp:
                    self.progress.force_to_finish_go()

                self._callback.put_nowait(SIGNAL_EMPTY_RECV(self))
                break
            self.progress.go(len(buff) - prv_len)

            if self.progress.go_inc >= self.progress.length:
                self._buffer(buff)
                break
            elif len(buff) >= 1048576:  # 1 MB
                self._buffer(buff)
                del buff
                buff = b''

        self.progress.__stop_handler()

    def close(self):
        response = self._response
        connection = self._connection
        self._response = None
        self._connection = None
        self.__connected = False

        if response:
            response.close()
        if connection:
            connection.sock.shutdown(socket.SHUT_RDWR)
            connection.close()

    def getheader(self, name, default=None):
        return self._response.getheader(name, default)

    def run(self, callback):
        self._callback = callback

        if not self.__signal.empty():
            signal = self.__signal.get()
            if signal.id == ID_TASK_SLICE:
                putrange = self.progress.slice_check(signal.content)
                self._callback.put_nowait(
                    SIGNAL_TASK_SLICE(Slice(client=HTTPClient, range=putrange))
                )
            elif signal.id == ID_SWITCH:
                self.source = signal.content
            elif signal.id == ID_TASK_PAUSE:
                self._callback = None
                callback.put_nowait(SIGNAL_THREAD_END(self))
            elif signal.id == ID_WAIT:
                sleep(signal.content)
            self.__signal.task_done()

        try:
            if not self.__connected:
                self.connect()
            self.retrieve()
        except (socket.gaierror, socket.timeout, URLUnknownError,
                HTTP4XXError, URLError, HTTPError, Exception) as e:
            self._exception_handler(e)
        finally:
            self.close()

        self._callback = None
        callback.put_nowait(SIGNAL_THREAD_END(self))

    def open_only(self, callback):
        self._callback = callback

        res = None
        try:
            res = self.connect()
        except (socket.gaierror, socket.timeout, URLUnknownError,
                HTTP4XXError, URLError, HTTPError, Exception) as e:
            self._exception_handler(e)

        self._callback = None
        return res

    def _exception_handler(self, exception):
        client_exc = ClientException(client=self, exception=exception)
        if type(exception) in (URLError, HTTPError, socket.timeout):
            signal_type = SIGNAL_TIMEOUT
        elif type(exception) is socket.gaierror:
            signal_type = SIGNAL_GAIERROR
        elif type(exception) is HTTP4XXError:
            signal_type = SIGNAL_CRASH
        else:
            signal_type = SIGNAL_UNKNOWN

        self._callback.put_nowait(
            SIGNAL_URL_STATUS(signal_type(client_exc))
        )

    def install_callback(self, callback):
        self._callback = callback

    def clear_callback(self):
        self._callback = None

    def send_signal(self, signal):
        self.__signal.put_nowait(signal)

    def write_to_file(self, fp):
        if not self.progress.is_empty_buff():
            fp.seek(self.progress.begin + self.progress.done_inc)
            fp.writelines(self.progress.fetch_buffer())

    def getprogress(self):
        return self.progress

    def getsource(self):
        return self.source

    def getresonse(self):
        return self._response

    def clear_signal(self):
        del self.__signal
        self.__signal = Queue()


def build_initial_opener(source, progress, callback):
    client = HTTPClient(source, progress)
    client.install_callback(callback)
    res = client.open_only(callback)

    if not res:
        raise URLConnectFailed()

    filesize = res.length
    filename = None

    cont = client.getheader('content-disposition')
    if cont:
        filename = re.search('filename="?(.*)"?', cont).group(1)

    if not filename:
        filename = source.path.split('/')[-1]

    if not filename:
        filename = None

    progress.range = (0, filesize)

    if not res.chunked and not filesize:
        raise AssertionError()

    client.clear_callback()
    return InitialResult(client=client, filename=filename, filesize=filesize or float('inf'),
                         unspecified_size=res.chunked, response=res)





from multiprocessing import Pipe, Process, current_process
from six.moves.queue import Queue
from threading import current_thread
from collections import namedtuple
from nbdler.misc.thread import ThreadCollector
from traceback import format_exc


class ProcessRuntimeError(Exception):
    pass


HandleSignal = namedtuple('HandleSignal', 'id name args kwargs')
ResponseSignal = namedtuple('ResponseSignal', 'id value traceback')

HandleSignal.__new__.__defaults__ = ((), None)
ResponseSignal.__new__.__defaults__ = (None, None)

TERMINATE_PROCESS_SIGNAL = HandleSignal(id=None, name='')
TERMINATE_RESPONSE_SIGNAL = ResponseSignal(id=None)


def make_method(func=None, block=True):
    if func is not None:
        def wrapper(self, *args, **kwargs):
            cur_t = current_thread()
            cur_queue = Queue()
            self._process_response_id_queue[cur_t.ident] = cur_queue
            h_signal = HandleSignal(id=cur_t.ident, name=func.__name__,
                                    args=args, kwargs=kwargs)
            self._parent_conn.send(h_signal)
            if block:
                res_signal = cur_queue.get()
                if res_signal.traceback:
                    raise res_signal.value

                return res_signal.value

        return wrapper
    else:
        def inner_wrapper(f):
            return make_method(f, block=block)

        return inner_wrapper


def make_class(handle_class):
    def wrapper(wrapped_class):
        class MultiprocessClassWrapper(wrapped_class):
            def __init__(self, *args, **kwargs):
                self._parent_conn, self.__child_conn = Pipe()
                self.__class_handle = handle_class
                self.__args = args
                self.__kwargs = kwargs
                self.__thread_collector = ThreadCollector()
                self._process_response_id_queue = {}
                self._process = Process(target=child_process_main_thread,
                                        args=(self.__child_conn, handle_class, args, kwargs))
                self._process.start()
                self.__thread_collector.put(self.__response_receiver_thread,
                                            '(pid_%d)C_recv' % current_process().pid).start()
                self._closed = False

            def __response_receiver_thread(self):
                while True:
                    signal = self._parent_conn.recv()
                    if signal.id is None:
                        break
                    self._process_response_id_queue[signal.id].put_nowait(signal)
                    if signal.traceback:
                        if type(signal.value) is ProcessRuntimeError:
                            self.__thread_collector.put(
                                self.close, '(pid_%d)P_close' % current_process().pid).start()

            def close(self, *args, **kwargs):
                if self._closed:
                    raise RuntimeError('(pid:%d)process has already closed.' % current_process().pid)
                self._closed = True
                try:
                    make_method(self.close)(self, *args, **kwargs)
                finally:
                    self._parent_conn.send(TERMINATE_PROCESS_SIGNAL)
                    self.__child_conn.send(TERMINATE_PROCESS_SIGNAL)
                    self._process.join()
                    self._process.close()

        return MultiprocessClassWrapper

    return wrapper


def child_process_main_thread(child_conn, class_handle, class_args=(), class_kwargs=None):
    def handle_thread(handle_signal):
        try:
            retval = getattr(class_instance, handle_signal.name)(
                *handle_signal.args, **handle_signal.kwargs)
        except Exception as e:
            ret_signal = ResponseSignal(id=handle_signal.id, value=e,
                                        traceback=format_exc())
        else:
            ret_signal = ResponseSignal(id=handle_signal.id, value=retval)

        response_queue.put_nowait(ret_signal)

    def send_response_thread():
        while True:
            res_signal = response_queue.get()
            if res_signal.id is None:
                break
            try:
                child_conn.send(res_signal)
            except Exception as e:
                res_signal = ResponseSignal(id=res_signal.id,
                                            value=ProcessRuntimeError(e),
                                            traceback=format_exc())
                child_conn.send(res_signal)

    def recv_handle_thread():
        while True:
            signal = child_conn.recv()
            if signal.id is None:
                response_queue.put_nowait(TERMINATE_RESPONSE_SIGNAL)
                break
            thread_collector.put(handle_thread, signal.name, args=(signal,)).start()

    thread_collector = ThreadCollector()
    class_instance = class_handle(*class_args, **class_kwargs)
    response_queue = Queue()
    cur_p = current_process()
    thread_collector.put(send_response_thread, '(pid_%d)C_send' % cur_p.pid).start()
    thread_collector.put(recv_handle_thread, '(pid_%d)C_recv' % cur_p.pid).start()

    thread_collector.join()






from nbdler.console.multi_client import MultiClientConsole
from nbdler.console.single_client import SingleClientConsole
from nbdler.struct.dump import HandlerDumpedData, FileDumpedData
from nbdler.file import File, FileInfo
from nbdler.client import build_initial_opener
from nbdler.struct.progress import Progress
from nbdler.exception import URLPoolEmpty, URLConnectFailed, \
    URLMaxRetriesExceeded, HandlerBusy, FileAlreadyExisted
from nbdler.request import Request
from nbdler.url.pool import Url
from nbdler.saver import Saver
from nbdler.misc.process_interface import make_method, make_class
from nbdler.misc.signal import ID_TASK_STOP, ID_TASK_EXCEPTION

from six.moves.queue import Queue
import os


DEFAULT_MAX_THREAD = 5
DEFAULT_BUFFSIZE = 10 * 1024 * 1024     # 20MB


def dlopen(source=None, *, maxthread=DEFAULT_MAX_THREAD, buffsize=DEFAULT_BUFFSIZE, child_process=False):
    if child_process:
        open_handler = HandlerForChildProcess
    else:
        open_handler = Handler

    return open_handler(source, max_thread=maxthread, buffsize=buffsize)


class Handler:
    """
        Attribute:
            source: A Source to make a Download Handler, should be a Download Request
                    or Handler datafile.

            max_thread: Specify the upper limit of running client at the same time.

            buffsize: Specify the max download buffer size keeping in memory.
    """

    __idle = True
    __new_file = True

    def __init__(self, source=None, *, max_thread=DEFAULT_MAX_THREAD, buffsize=DEFAULT_BUFFSIZE):

        self.url = Url()
        self._console = None

        self._max_thread = max_thread
        self._buffsize = buffsize
        self._saver = None
        self._inner_exc_signal = Queue()

        if source:
            self.open(source)

    def puturl(self, url, headers=None, cookie=None, *,
               proxy=None, max_conn=None, rangef=None, name=None):
        self.url.put(url=url, headers=headers, cookie=cookie, proxy=proxy,
                     max_conn=max_conn, rangef=rangef, name=name)

    def geturlinfo(self, k, default=None):
        srcwrapper = self.url.get(k, default)
        return srcwrapper.source.getinfo()

    def geturlinfo_all(self):
        retinfos = []
        for i, v in enumerate(self.url.get_all()):
            if v:
                retinfos.append(self.geturlinfo(i))

        return retinfos

    def open(self, source):
        """ Make a decision to open Download Handler from a request or a datafile. """
        if not self.is_idle():
            raise HandlerBusy('handler has already opened.')
        if type(source) is Request:
            self._open_request(source)
        elif type(source) is str:
            self._open_file(source)
        else:
            raise AssertionError()
        # return self

    def _open_file(self, source):
        """ Open Download Handler from a Download Handler datafile. """
        self.__new_file = False
        self.__idle = False
        dumped_data = Saver.json_loads(source)
        console = self._load_console(dumped_data)

        self._saver = Saver(os.path.join(console.file.getpath(), console.file.getname() + '.nb'),
                            self.save_handler)
        self._console.install_saver(self._saver)
        self.__idle = True

    def _urlopen_console(self, dlrequest):
        self.url.open_request(dlrequest)
        url_id = 0
        srcwrap = self.url.get(url_id)
        srcurl = srcwrap.get()
        if not srcwrap:
            raise URLPoolEmpty()

        lprog = Progress((0,))

        callback = Queue()
        max_retried = dlrequest.max_retries
        while True:
            try:
                init_res = build_initial_opener(srcurl, lprog, callback)
            except URLConnectFailed:
                if self.is_idle():
                    return None
                if max_retried is None:
                    continue

                max_retried -= 1
                if max_retried <= 0:
                    raise URLMaxRetriesExceeded()
            else:
                break

        specify_path, specify_name = os.path.split(dlrequest.filepath)
        if not specify_name:
            specify_name = init_res.filename
        file = File(specify_path, specify_name, init_res.filesize)
        if os.path.exists(os.path.join(specify_path, specify_name)):
            raise FileAlreadyExisted()

        srcwrap.use()

        if init_res.unspecified_size:
            console_handler = SingleClientConsole
            blocksize = float('inf')
        else:
            console_handler = MultiClientConsole
            blocksize = dlrequest.block_size

        srcwrap.source.response(init_res.client.geturl(),
                                init_res.response.getheaders(),
                                init_res.response.getcode(),
                                init_res.filesize)
        console = console_handler(file, self.url, self._buffsize,
                                  blocksize, dlrequest.max_thread or self._max_thread, self._inner_exc_signal)
        console.insert(init_res.client, lprog)
        return console

    def _open_request(self, dlrequest):
        """ Open Download Handler from a Download Request. """
        self.__new_file = True
        self.__idle = False
        if not dlrequest.resume:
            console = self._urlopen_console(dlrequest)
            if console:
                self._saver = Saver(os.path.join(console.file.getpath(), console.file.getname() + '.nb'),
                                    self.save_handler)
                console.install_saver(self._saver)
                self._console = console

            self.__idle = True
        else:
            self._open_file(dlrequest.filepath + '.nb')
            # dumped_data = Saver.json_loads(dlrequest)
            # console = self._load_console(dumped_data)
            for i in dlrequest:
                self.url.put(i.url, i.headers, i.cookie, proxy=i.proxy, max_conn=i.max_thread,
                             rangef=i.rangef, name=i.name)


    def getavgspeed(self):
        """ Return download average speed. """
        return self._console.getavgspeed()

    def getinstspeed(self):
        """ Return download instant speed.

        It gets instant speed from a sampling thread which
        sampling at INST_SPEED_REFRESH_FREQUENCY Hz, default 2 Hz.
        """
        return self._console.getinstspeed()

    def getincbyte(self):
        """ Return the size of download bytes already. """
        return self._console.getincbyte()

    def get_remain_time(self):
        """ Return the time remaining to finish the task. (base on average speed)"""
        return self._console.get_remain_time()

    def get_remain_byte(self):
        """ Return the remaining size of downloading file. """
        return self._console.get_remain_byte()

    def save_handler(self):
        """ Save Download Handler Data to the file name with a suffix of '.nb'. """
        self._saver.json_dumps(tuple(self._dump_data()))

    def _dump_data(self):
        """ Return a raw Download Handler data to serialize. """
        return HandlerDumpedData(url=tuple(self.url.dump_data()), file=tuple(self._console.file.dump_data()),
                                 console=tuple(self._console.dump_data()), maxthread=self._max_thread,
                                 buffsize=self._buffsize, blocksize=self._console.getblocksize())

    def _load_console(self, dumped_data):
        """ Load Download Handler from a json serializing Handler data. """
        data = HandlerDumpedData(*dumped_data)
        file_data = FileDumpedData(*data.file)

        file = File(file_data.path, file_data.name, file_data.size)
        self.url.load(data.url)
        self._max_thread = data.maxthread
        self._buffsize = data.buffsize

        if file.getsize():
            console_handler = MultiClientConsole
        else:
            console_handler = SingleClientConsole
        console = console_handler(file, self.url, data.buffsize,
                                  data.blocksize, data.maxthread,
                                  self._inner_exc_signal)
        console.load(data.console)

        return console

    def getfileinfo(self):
        """ Return a static copy of the download file's information. """
        return FileInfo(name=self._console.file.getname(),
                        path=self._console.file.getpath(),
                        size=self._console.file.getsize(),
                        block_size=self._console.getblocksize())

    def start(self):
        """ Start the download task. """
        if not self.is_idle():
            raise RuntimeError('cannot start an busy handler.')

        if self.is_running():
            raise RuntimeError('cannot start a started handler.')

        if not self.is_opened():
            raise RuntimeError('cannot start a unopened handler.')

        self.__idle = False

        if self._console.file.getsize() and self.__new_file:
            self._console.file.makefile()
            self.__new_file = False
            self.save_handler()

        self._console.prepare()
        self._console.run()

    def pause(self, block=True):
        """ Stop the running/opening download task. """
        if self._console:
            self._console.pause(block)
        self.__idle = True

    stop = pause

    def is_finish(self):
        """ Return True if download task is finished. """
        return self._console.is_finish()

    def is_running(self):
        """ Return True if download task is started. """
        return self._console is not None and self._console.runflag

    def get_online_counter(self):
        """ Return the number of running client threads. """
        return self._console.get_online_counter()

    def close(self):
        """ Remove the finished download task's datafile. """
        if self.is_finish():
            datafile = os.path.join(self._console.file.getpath(), self._console.file.getname() + '.nb')
            if os.path.exists(datafile):
                os.unlink(datafile)
        else:
            raise RuntimeError("cannot close an unfinished task.")

    def join(self, timeout=None):
        """ Blocks the calling thread until the end of download or getting a download stop. """
        self._console.join(timeout)

    def trap(self):
        """ A inner-exception sensitive join(). """
        if self.is_idle():
            raise RuntimeError('cannot trap a download handler when it is\'t started')
        while True:
            signal = self._inner_exc_signal.get()
            if signal.id == ID_TASK_EXCEPTION:
                raise signal.content
            elif signal.id == ID_TASK_STOP:
                break

    def install_external_callback_queue(self, callback_handle):
        """ Install a callback signal queue.
         when task is finished or stopped, there is a signal
         sent to callback_handle (should be a type of CallbackHandle).
         """
        self._console.install_external_callback_queue(callback_handle)

    def is_opened(self):
        return self._console is not None

    def is_idle(self):
        if self._console:
            return not self._console.runflag
        else:
            return self.__idle


@make_class(Handler)
class HandlerForChildProcess:

    @make_method
    def start(self):
        pass

    @make_method
    def pause(self, block=False):
        pass

    @make_method
    def open(self, source):
        pass

    @make_method
    def join(self, timeout=None):
        pass

    @make_method
    def trap(self):
        pass

    @make_method
    def install_external_callback_queue(self, callback_handle):
        pass

    @make_method
    def is_opened(self):
        pass

    @make_method
    def is_idle(self):
        pass

    @make_method
    def close(self):
        pass

    @make_method
    def get_online_counter(self):
        pass

    @make_method
    def is_running(self):
        pass

    @make_method
    def is_finish(self):
        pass

    @make_method
    def getfileinfo(self):
        pass

    @make_method
    def getavgspeed(self):
        pass

    @make_method
    def getinstspeed(self):
        pass

    @make_method
    def getincbyte(self):
        pass

    @make_method
    def get_remain_time(self):
        pass

    @make_method
    def get_remain_byte(self):
        pass

    @make_method
    def save_handler(self):
        pass

    @make_method
    def puturl(self, url, headers=None, cookie=None, *,
               proxy=None, mxthread=None, rangef=None, name=None):
        pass

    @make_method
    def geturlinfo(self, k, default=None):
        pass

    @make_method
    def geturlinfo_all(self):
        pass

    # @make_method
    # def rmurl(self):







from six import add_metaclass

from abc import ABCMeta, abstractmethod


@add_metaclass(ABCMeta)
class AbstractClient:

    @classmethod
    @abstractmethod
    def connect(cls):
        pass

    @classmethod
    @abstractmethod
    def retrieve(cls):
        pass

    @classmethod
    @abstractmethod
    def send_signal(cls, signal):
        pass

    @classmethod
    @abstractmethod
    def write_to_file(cls, fp):
        pass

    @classmethod
    @abstractmethod
    def install_callback(cls, callback):
        pass

    @classmethod
    @abstractmethod
    def clear_callback(cls):
        pass


@add_metaclass(ABCMeta)
class AbstractResponse:

    @classmethod
    @abstractmethod
    def read(cls):
        pass






from threading import Thread as _Thread, current_thread
from threading import Lock, RLock
from time import perf_counter


class Thread(_Thread):
    __slots__ = ('_owner',)

    def __init__(self, owner=None, *args, **kwargs):
        _Thread.__init__(self, *args, **kwargs)
        self._owner = owner

    def is_stopped(self):
        """ Return True if thread is stopped. """
        return not self.is_alive() and self._is_stopped

    def is_started(self):
        """ Return True if thread is started. """
        return self._started.is_set()

    @property
    def owner(self):
        """ Return thread's owner. """
        return self._owner


class ThreadCollector:
    def __init__(self, daemon=False):
        self._running_threads = {}
        self.__make_lock = Lock()
        self._daemon = daemon

    @property
    def daemon(self):
        return self._daemon

    def put(self, target, group_name, args=(), owner=None, kwargs=None):
        """ Make a new thread into thread pool. """
        self._correct_running()
        target = Thread(target=target, name=group_name, args=args,
                        daemon=self._daemon, owner=owner, kwargs=kwargs)
        with self.__make_lock:
            if group_name not in self._running_threads:
                self._running_threads[group_name] = []

            self._running_threads[group_name].append(target)
        return target

    def wait(self, group_name, timeout=None):
        """ Wait a specified group of threads until to the end of life. """
        cur_thread = current_thread()
        for thread in self._running_threads.get(group_name, []):
            if thread != cur_thread:
                timeout = _join_one(thread, timeout)

    def join(self, timeout=None):
        """ Wait all the threads in the pool until to the end of life. """
        cur_thread = current_thread()
        for v in list(self._running_threads.values()):
            for thread in v:
                if thread != cur_thread:
                    timeout = _join_one(thread, timeout)

    def start_all(self):
        """ Start all the threads in the pool. """
        for v in list(self._running_threads.values()):
            for thread in v:
                if not thread.is_started() and not thread.is_stopped():
                    thread.start()

    def start_group(self, name):
        """ Start a specified group of threads to run. """
        for thread in self._running_threads[name]:
            if not thread.is_started() and not thread.is_stopped():
                thread.start()

    def get_group(self, group_name):
        """ Return all threads in a specified group. """
        self._correct_running()
        return self._running_threads.get(group_name, [])

    def _correct_running(self):
        """ Clean all the stopped thread. """
        with self.__make_lock:
            for i, v in list(self._running_threads.items()):
                for thread in v:
                    if thread.is_stopped():
                        v.remove(thread)
                if not v:
                    del self._running_threads[i]

    def get_all(self):
        """ Return all running threads. """
        self._correct_running()
        return self._running_threads


def _join_one(thread, timeout=None):
    if timeout is not None and timeout < 0:
        raise TimeoutError()
    start = perf_counter()
    thread.join(timeout=timeout)
    if timeout is not None:
        timeout -= perf_counter() - start

    return timeout


# _GLOBAL_THREAD_POOLS = {}
#
#
# def get(name, *, owner=None, daemon=False):
#     global _GLOBAL_THREAD_POOLS
#     if name not in _GLOBAL_THREAD_POOLS:
#         _GLOBAL_THREAD_POOLS[name] = ThreadCollector(daemon)
#
#     if owner is not None:
#         name = name + '<%d>' % id(owner)
#
#     return _GLOBAL_THREAD_POOLS[name]
#
#
# def remove(name):
#     global _GLOBAL_THREAD_POOLS
#     if name not in _GLOBAL_THREAD_POOLS:
#         raise ValueError('thread pool named %s isn\'t existed.' % name)
#
#     del _GLOBAL_THREAD_POOLS[name]
#
#
# def has_key(name):
#     global _GLOBAL_THREAD_POOLS
#     return name in _GLOBAL_THREAD_POOLS





import io
import os
from collections import namedtuple
from typing import AnyStr, List
from nbdler.struct.dump import FileDumpedData

FileInfo = namedtuple('FileInfo', 'name path size block_size')


class File:
    def __init__(self, path, name, size):
        self._name = name
        self._path = path
        self._size = size

        self._fp = None
        self._closed = True

    @property
    def ext(self):
        return os.path.splitext(self._name)[-1]

    def getpath(self):
        return self._path

    def getname(self):
        return self._name

    def getsize(self):
        return self._size

    def open(self):
        fp = io.open(os.path.join(self._path, self._name), mode='rb+')
        self._fp = fp
        self._closed = False
        return self

    def makefile(self):
        with io.open(os.path.join(self._path, self._name), mode='wb') as fp:
            if self._size != float('inf'):
                fp.seek(self._size - 1)
                fp.write(b'\x00')

    def write(self, s: AnyStr) -> int:
        return self._fp.write(s)

    def writelines(self, lines: List[AnyStr]) -> None:
        self._fp.writelines(lines)

    def close(self) -> None:
        if self._fp is not None:
            self._fp.close()
            self._fp = None
            self._closed = True

    def isclosed(self):
        return self._closed

    def __del__(self):
        self.close()

    def seek(self, offset: int, whence: int = 0) -> int:
        return self._fp.seek(offset, whence)

    def flush(self) -> None:
        self._fp.flush()

    def dump_data(self):
        return FileDumpedData(path=self._path, name=self._name, size=self._size)






import gzip
import io
import json


class Saver:
    def __init__(self, saved_path, save_handle):
        self._saved_path = saved_path
        self._save_handle = save_handle

    def json_dumps(self, obj):
        with io.open(self._saved_path, 'wb') as fp:
            fp.write(gzip.compress(json.dumps(obj).encode('utf-8')))
            fp.flush()

    def dump(self):
        self.json_dumps(self._save_handle())

    @staticmethod
    def json_loads(file):
        with io.open(file, 'rb') as f:
            json_byte = gzip.decompress(f.read())

        return json.loads(json_byte, encoding='utf-8')






from nbdler.url.basic import BasicUrl
from nbdler.struct.dump import UrlResponseDumpedData


class UrlResponse(BasicUrl):
    def __init__(self, url, headers, code, length):
        super(UrlResponse, self).__init__(url, headers)
        self.code = code
        self.length = length

    def dump_data(self):
        return UrlResponseDumpedData(url=self.url, headers=dict(self.headers),
                                     code=self.code, length=self.length)





from six import add_metaclass
from abc import abstractmethod, ABCMeta


@add_metaclass(ABCMeta)
class AbstractConsole:

    @abstractmethod
    def prepare(self):
        pass

    @abstractmethod
    def run(self):
        """ Run download task. """

    @abstractmethod
    def pause(self, *args, **kwargs):
        """ Pause downloading. """

    @abstractmethod
    def insert(self, *args, **kwargs):
        """ Insert a block described by Client and Progress into the console. """

    @abstractmethod
    def getincbyte(self):
        """ Return byte size downloaded. """

    @abstractmethod
    def getinstspeed(self):
        """ Return instant download speed. """

    @abstractmethod
    def getavgspeed(self):
        """ Return average download speed. """

    @abstractmethod
    def get_remain_byte(self):
        """ Return remaining byte to finish the download task. """

    @abstractmethod
    def get_remain_time(self):
        """ Return remaining time to finish the download task. """

    @abstractmethod
    def is_finish(self):
        """ Return True if download task finished. """

    @abstractmethod
    def _buffer_release_thread(self):
        """ A thread handler about writing buffer to file. """

    @abstractmethod
    def _client_callback_handle_thread(self):
        """ A thread handler about callback signal. """

    @abstractmethod
    def getblocksize(self):
        """ Return console's unit block size. """




from collections import namedtuple
from wsgiref.headers import Headers
from six.moves.urllib.parse import urlparse

UrlInfo = namedtuple('UrlInfo', 'url headers cookie proxy max_conn rangef name response')
UrlInfo.__new__.__defaults__ = (None,)


class BasicUrl:
    def __init__(self, url, headers):
        self._url = None
        self._urlparse = None

        if headers is None:
            headers = Headers([])
        elif isinstance(headers, dict):
            headers = Headers(list(headers.items()))
        elif isinstance(headers, list):
            headers = Headers(headers)
        else:
            raise TypeError()
        self.headers = headers
        self.url = url

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, value):
        self._url = value
        self._urlparse = urlparse(value)

    @property
    def hostname(self):
        if not self._urlparse:
            return None
        return self._urlparse.hostname

    @property
    def port(self):
        if not self._urlparse:
            return None
        return self._urlparse.port

    @property
    def path(self):
        if not self._urlparse:
            return None
        return self._urlparse.path

    @property
    def scheme(self):
        if not self._urlparse:
            return None
        return self._urlparse.scheme

    @property
    def query(self):
        if not self._urlparse:
            return None
        return self._urlparse.query

    @property
    def netloc(self):
        if not self._urlparse:
            return None
        return self._urlparse.netloc






class Error(Exception):
    pass


# class ProcessRuntimeError(Error):
#     pass

class ManagerError(Error):
    pass


class ChildProcessDisable(ManagerError):
    pass


class HandlerError(Error):
    pass


class ClientError(Error):
    pass


class FileAlreadyExisted(HandlerError):
    pass


class HandlerBusy(HandlerError):
    pass


class HTTP4XXError(ClientError):
    pass


class UnknownError(HandlerError):
    pass


class URLError(Error):
    pass


class URLPoolEmpty(URLError):
    pass


class InvalidURLError(URLError):
    pass


class URLCrash(URLError):
    pass


class URLTimeout(URLError):
    pass


class URLUnknownError(URLError):
    pass


class NetworkDisconnected(URLError):
    pass


class MaxUsedExceedError(URLError):
    pass


class URLConnectFailed(URLError):
    pass


class URLMaxRetriesExceeded(URLError):
    pass












from setuptools import setup, find_packages
import io

version = '0.9.9'
author = 'ZSAIm'
author_email = 'zzsaim@163.com'

with io.open('README.rst', 'r', encoding='utf-8') as freadme:
    long_description = freadme.read()

setup(
    name='Nbdler',
    version=version,
    description='HTTP/HTTPS downloader',
    long_description=long_description,
    author=author,
    author_email=author_email,
    url='https://github.com/ZSAIm/Nbdler',
    license='Apache-2.0 License',
    classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: Apache Software License',
            'Programming Language :: Python',
            "Programming Language :: Python :: 2",
            "Programming Language :: Python :: 2.7",
            "Programming Language :: Python :: 3",
            'Programming Language :: Python :: 3.3',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
        ],
    packages=find_packages(),

)